import fnmatch
import ujson

from os import listdir, path
from . import eventlog

_UMMR0A_JNODE_FILENAME = 'ummr0a-type29-jnode.json'
BASE_SPOOL_DIRPATH = '/var/spool/edge/'

BASE_BACKUP_DIRPATH = '/etc/edge/setupstore/backup/'
BASE_EDGE_JNODE_DIRPATH = '/etc/edge/'

_BASE_UI_JNODE_DIRPATH = '/etc/edge/setupstore/admin/actions/'
UI_TIME_JNODE_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'time.json')
BASE_UI_PIPELINE_JNODE_DIRPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'pipeline')
DB_JSON_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'db.json')
PSUCTL_JSON_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'psu-ctl.json')

ABOUT_JSON_FILEPATH = '/etc/edge/setupstore/about.json'

_BASE_UI_SCHEMA_DIRPATH = '/etc/edge/setupstore/schemas/'
BASE_UI_PIPELINE_SCHEMA_DIRPATH = path.join(_BASE_UI_SCHEMA_DIRPATH, 'pipeline')

SHARE_UI_DIRPATH = '/opt/edge/share/setupstore/ui/'
SHARE_PIPELINE_DIRPATH = path.join(SHARE_UI_DIRPATH, 'pipeline')

NN_CONF_FILEPATH = '/etc/edge-nn/main.conf'

EDGE_TYPE_SETUPSTORE = 'setupstore'
EDGE_TYPE_EVENTSTORE = 'eventstore'
EDGE_TYPE_LPR = 'lpr'
EDGE_TYPE_MEDIACODEC = 'mediacodec'
EDGE_TYPE_MEDIAGRABBER = 'mediagrabber'
EDGE_TYPE_GRABCONTROL = 'grabcontrol'
EDGE_TYPE_RADAR = 'radar'
EDGE_TYPE_RENDERER = 'overlayrender'
EDGE_TYPE_RTSP_GRABBER = 'rtspgrabber'
EDGE_TYPE_RTSP_SERVER = 'rtspserver'
EDGE_TYPE_SPEEDCONTROL = 'speedcontrol'
EDGE_TYPE_TRACKER = 'tracker'
EDGE_TYPE_WEBSERVER = 'webserver'
EDGE_TYPE_RAWCODEC = 'rawcodecgst'
EDGE_TYPE_BLOBSTORE = 'blobstore'
EDGE_TYPE_MEDIACODEC_GST = 'mediacodecgst'
EDGE_TYPE_IMAGE_MAPPER = 'imagemapper'
EDGE_TYPE_TRAFFIC_LIGHT = 'traffic-light'
EDGE_TYPE_LBSAPP = 'lbsapp'

SETUPSTORE_CONF_FILEPATH = path.join(BASE_EDGE_JNODE_DIRPATH, EDGE_TYPE_SETUPSTORE, 'main.conf')
BLOBSTORE_CONF_FILEPATH = path.join(BASE_EDGE_JNODE_DIRPATH, EDGE_TYPE_BLOBSTORE + '.65535', 'main.conf')

SYSTEST_CONF_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'systest', '0')
SYSTEST_REPORTS_DIRPATH = path.join(_BASE_UI_JNODE_DIRPATH, 'systest', 'reports')

SIGN_CONF_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, "sign", "0.json")
SIGN_KEY_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, "sign", "sign.key")
SIGN_CRT_FILEPATH = path.join(_BASE_UI_JNODE_DIRPATH, "sign", "sign.crt")
SIGN_DEFAULT_CONF_FILEPATH = path.join(SHARE_UI_DIRPATH, "sign-jnode.json")

PTYPE_LPR_RTSP = '1'  # rtsp-videosource; radar absent
PTYPE_SPEED = '2'     # gige-or-usb-videosource; radar present
PTYPE_LPR_MV = '3'     # gige-or-usb-videosource; radar absent

UI_MODULES = 'modules'
UI_TYPE_PIPELINE = ''
UI_TYPE_EVENTSTORE = 'eventstore'
UI_TYPE_GEOLOCATION = 'geolocation'
UI_TYPE_HTTP_STREAMING = 'http-streaming'
UI_TYPE_LANE_ENFORCEMENT = 'lane-enforcement'
UI_TYPE_LENS = 'lens'
UI_TYPE_LPR = 'lpr'
UI_TYPE_LOCATION = 'location'
UI_TYPE_RTSP_STREAMING = 'rtsp-streaming'
UI_TYPE_VIDEOSOURCE = 'videosource'

UI_TYPE_SIGN = 'sign'

UI_DETECTORS = 'detectors'
UI_TYPE_TRANSIT = 'transit'
UI_TYPE_PARKING = 'parking'
UI_TYPE_SPOT_SPEED = 'spot-speed'
UI_TYPE_SECTION_SPEED = 'section-speed'
UI_TYPE_WRONG_WAY = 'wrong-way'
UI_TYPE_WANTED = 'wanted'
UI_TYPE_RED_LIGHT = 'red-light'


class UiIdentity:
    def __init__(self, ui_ptype, ui_type, ui_id):
        """
        :type ui_ptype: str
        :type ui_type: str
        :type ui_id: str
        """
        self.ptype = ui_ptype
        self.type = ui_type
        self.id = ui_id

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'ptype', None) == self.ptype and
                getattr(other, 'id', None) == self.id)

    def __hash__(self):
        return hash(self.ptype + self.type + self.id)


class EdgeIdentity:
    def __init__(self, edge_type, edge_gid):
        """
        :type edge_type: str
        :type edge_gid: str
        """
        self.type = edge_type
        self.gid = edge_gid

    def __hash__(self):
        return hash(self.type + self.gid)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'gid', None) == self.gid)


# noinspection PyClassHasNoInit
class Pipeline:
    # noinspection PyClassHasNoInit
    class Speed:
        # noinspection PyClassHasNoInit
        class Ui:
            PIPELINE = UiIdentity(PTYPE_SPEED, UI_TYPE_PIPELINE, '')
            GEOLOCATION_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_GEOLOCATION, '0')

            VIDEOSOURCE_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_VIDEOSOURCE, '0')
            LENS_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_LENS, '0')

            LPR_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_LPR, '0')
            LANE_ENFORCEMENT_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_LANE_ENFORCEMENT, '0')

            LOCATION_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_LOCATION, '0')

            EVENTSTORE_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_EVENTSTORE, '0')
            TRANSIT_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_TRANSIT, '0')

            HTTP_STREAMING_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_HTTP_STREAMING, '0')
            RTSP_STREAMING_0 = UiIdentity(PTYPE_SPEED, UI_TYPE_RTSP_STREAMING, '0')

        # noinspection PyClassHasNoInit
        class Edge:
            SETUPSTORE_0 = EdgeIdentity(EDGE_TYPE_SETUPSTORE, '0')

            MEDIAGRABBER_0 = EdgeIdentity(EDGE_TYPE_MEDIAGRABBER, '0')
            GRAB_CONTROL_0 = EdgeIdentity(EDGE_TYPE_GRABCONTROL, '0')
            RAW_CODEC_0 = EdgeIdentity(EDGE_TYPE_RAWCODEC, '0')

            RADAR_0 = EdgeIdentity(EDGE_TYPE_RADAR, '0')
            LPR_0 = EdgeIdentity(EDGE_TYPE_LPR, '0')

            TRACKER_0 = EdgeIdentity(EDGE_TYPE_TRACKER, '0')

            SPEEDCONTROL_0 = EdgeIdentity(EDGE_TYPE_SPEEDCONTROL, '0')
            EVENTSTORE_65535 = EdgeIdentity(EDGE_TYPE_EVENTSTORE, '65535')

            RENDERER_0 = EdgeIdentity(EDGE_TYPE_RENDERER, '0')
            RENDERER_1 = EdgeIdentity(EDGE_TYPE_RENDERER, '1')

            WEBSERVER_0 = EdgeIdentity(EDGE_TYPE_WEBSERVER, '0')
            RTSPSERVER_0 = EdgeIdentity(EDGE_TYPE_RTSP_SERVER, '0')
            MEDIA_CODEC_GST_0 = EdgeIdentity(EDGE_TYPE_MEDIACODEC_GST, '0')
            BLOBSTORE_0 = EdgeIdentity(EDGE_TYPE_BLOBSTORE, '0')
            IMAGEMAPPER_0 = EdgeIdentity(EDGE_TYPE_IMAGE_MAPPER, '0')
            TRAFFIC_LIGHT_0 = EdgeIdentity(EDGE_TYPE_TRAFFIC_LIGHT, '0')
            LBSAPP_0 = EdgeIdentity(EDGE_TYPE_LBSAPP, '0')

    # noinspection PyClassHasNoInit
    class Rtsp:
        # noinspection PyClassHasNoInit
        class Ui:
            PIPELINE = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_PIPELINE, '')
            GEOLOCATION_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_GEOLOCATION, '0')
            VIDEOSOURCE_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_VIDEOSOURCE, '0')
            LPR_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_LPR, '0')
            LANE_ENFORCEMENT_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_LANE_ENFORCEMENT, '0')
            EVENTSTORE_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_EVENTSTORE, '0')
            TRANSIT_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_TRANSIT, '0')
            HTTP_STREAMING_0 = UiIdentity(PTYPE_LPR_RTSP, UI_TYPE_HTTP_STREAMING, '0')

        # noinspection PyClassHasNoInit
        class Edge:
            SETUPSTORE_0 = EdgeIdentity(EDGE_TYPE_SETUPSTORE, '0')

            RTSPGRABBER_0 = EdgeIdentity(EDGE_TYPE_RTSP_GRABBER, '0')

            LPR_0 = EdgeIdentity(EDGE_TYPE_LPR, '0')

            TRACKER_0 = EdgeIdentity(EDGE_TYPE_TRACKER, '0')

            SPEEDCONTROL_0 = EdgeIdentity(EDGE_TYPE_SPEEDCONTROL, '0')
            EVENTSTORE_65535 = EdgeIdentity(EDGE_TYPE_EVENTSTORE, '65535')

            RENDERER_0 = EdgeIdentity(EDGE_TYPE_RENDERER, '0')

            WEBSERVER_0 = EdgeIdentity(EDGE_TYPE_WEBSERVER, '0')
            MEDIA_CODEC_GST_0 = EdgeIdentity(EDGE_TYPE_MEDIACODEC_GST, '0')
            BLOBSTORE_0 = EdgeIdentity(EDGE_TYPE_BLOBSTORE, '0')
            TRAFFIC_LIGHT_0 = EdgeIdentity(EDGE_TYPE_TRAFFIC_LIGHT, '0')
            LBSAPP_0 = EdgeIdentity(EDGE_TYPE_LBSAPP, '0')

    # noinspection PyClassHasNoInit
    class Mv:
        # noinspection PyClassHasNoInit
        class Ui:
            PIPELINE = UiIdentity(PTYPE_LPR_MV, UI_TYPE_PIPELINE, '')
            GEOLOCATION_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_GEOLOCATION, '0')
            VIDEOSOURCE_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_VIDEOSOURCE, '0')
            LENS_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_LENS, '0')
            LPR_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_LPR, '0')
            LANE_ENFORCEMENT_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_LANE_ENFORCEMENT, '0')
            EVENTSTORE_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_EVENTSTORE, '0')
            TRANSIT_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_TRANSIT, '0')
            HTTP_STREAMING_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_HTTP_STREAMING, '0')
            RTSP_STREAMING_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_RTSP_STREAMING, '0')

            LOCATION_0 = UiIdentity(PTYPE_LPR_MV, UI_TYPE_LOCATION, '0')

        # noinspection PyClassHasNoInit
        class Edge:
            SETUPSTORE_0 = EdgeIdentity(EDGE_TYPE_SETUPSTORE, '0')

            MEDIAGRABBER_0 = EdgeIdentity(EDGE_TYPE_MEDIAGRABBER, '0')
            GRAB_CONTROL_0 = EdgeIdentity(EDGE_TYPE_GRABCONTROL, '0')
            RAW_CODEC_0 = EdgeIdentity(EDGE_TYPE_RAWCODEC, '0')

            LPR_0 = EdgeIdentity(EDGE_TYPE_LPR, '0')

            TRACKER_0 = EdgeIdentity(EDGE_TYPE_TRACKER, '0')

            SPEEDCONTROL_0 = EdgeIdentity(EDGE_TYPE_SPEEDCONTROL, '0')
            EVENTSTORE_65535 = EdgeIdentity(EDGE_TYPE_EVENTSTORE, '65535')

            RENDERER_0 = EdgeIdentity(EDGE_TYPE_RENDERER, '0')
            RENDERER_1 = EdgeIdentity(EDGE_TYPE_RENDERER, '1')

            WEBSERVER_0 = EdgeIdentity(EDGE_TYPE_WEBSERVER, '0')
            RTSPSERVER_0 = EdgeIdentity(EDGE_TYPE_RTSP_SERVER, '0')
            MEDIA_CODEC_GST_0 = EdgeIdentity(EDGE_TYPE_MEDIACODEC_GST, '0')
            BLOBSTORE_0 = EdgeIdentity(EDGE_TYPE_BLOBSTORE, '0')
            IMAGEMAPPER_0 = EdgeIdentity(EDGE_TYPE_IMAGE_MAPPER, '0')
            TRAFFIC_LIGHT_0 = EdgeIdentity(EDGE_TYPE_TRAFFIC_LIGHT, '0')
            LBSAPP_0 = EdgeIdentity(EDGE_TYPE_LBSAPP, '0')


def get_edge_jnode_dirpath(edge_type, edge_id):
    """
    :type edge_type: str
    :type edge_id: str
    :rtype: str
    """
    return path.join(BASE_EDGE_JNODE_DIRPATH, '%s.%s' % (edge_type, edge_id))


def get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id):
    """
    :type ui_pid: str
    :type ui_type: str
    :type ui_id: str
    :rtype: str
    """
    return path.join(BASE_UI_PIPELINE_JNODE_DIRPATH, ui_pid, ui_type, '%s.json' % ui_id)


def get_ui_pipeline_jnode_dirpath(ui_pid, ui_identity):
    """
    :type ui_pid: str
    :type ui_identity: UiIdentity
    :rtype: str
    """
    return path.join(BASE_UI_PIPELINE_JNODE_DIRPATH, ui_pid, ui_identity.type)


def get_ui_def_jnode_filepath(ui_identity, raw_body):
    """
    :type ui_identity: UiIdentity
    :type raw_body: str | None
    :rtype: str
    """
    # rtsp
    if ui_identity == Pipeline.Rtsp.Ui.VIDEOSOURCE_0:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'rtsp-jnode.json')

    elif ui_identity == Pipeline.Rtsp.Ui.HTTP_STREAMING_0:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'lpr-jnode.json')

    # mv
    elif ui_identity in [Pipeline.Mv.Ui.HTTP_STREAMING_0, Pipeline.Mv.Ui.RTSP_STREAMING_0]:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'lpr-jnode.json')

    # speed
    elif ui_identity == Pipeline.Speed.Ui.LOCATION_0:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, _UMMR0A_JNODE_FILENAME)

    # elif ui_identity == Pipeline.Speed.Ui.MEDIASTORE_0:
    #     return join_path(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'short-long-term-jnode.json')

    elif ui_identity in [Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.HTTP_STREAMING_0]:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'speed-jnode.json')

    elif ui_identity in [Pipeline.Speed.Ui.TRANSIT_0]:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'def-speed-jnode.json')

    # videosource for mv cameras
    elif ui_identity in [Pipeline.Mv.Ui.VIDEOSOURCE_0, Pipeline.Speed.Ui.VIDEOSOURCE_0]:
        args = ujson.loads(raw_body)
        vendor = args['vendor']
        raw_model = str(args['model'])
        # to find file we need to replace space to "-"
        model = str.replace(raw_model, ' ', '-')
        # serial = args['serial']
        filelist = listdir(path.join(SHARE_PIPELINE_DIRPATH, UI_TYPE_VIDEOSOURCE))
        fname = fnmatch.filter(filelist, '%s_%s_*jnode.json' % (vendor, model))
        if fname.__len__() == 0:
            eventlog.log_d('fail: find default jnode for videosource: "%s"' % model)
            raise NotImplementedError('camera not supported')
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, fname[0])

    else:
        return path.join(SHARE_PIPELINE_DIRPATH, ui_identity.type, 'def-jnode.json')


def get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id):
    """
    :type ui_pid: str
    :type ui_type: str
    :type ui_id: str
    :rtype: str
    """
    return path.join(BASE_UI_PIPELINE_SCHEMA_DIRPATH, ui_pid, ui_type, '%s.json' % ui_id)


def get_spool_dir(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return path.join(BASE_SPOOL_DIRPATH, 'pipeline-%s' % ui_pid)


def get_setupstore_conf_text(edge_id, active=True):
    """
    :type edge_id: str
    :type active: bool
    :rtype: str
    """
    if active:
        return 'Active=true\nArgSet=--id=%s --log-level=info\n' % edge_id
    else:
        return 'Active=false\nArgSet=--id=%s --log-level=info\n' % edge_id


def get_origin_about_filepath(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return path.join(BASE_UI_PIPELINE_JNODE_DIRPATH, ui_pid, 'about.json')
