import os
import errno
import signal
import shutil
import subprocess
import ujson

from . import eventlog
from . import path_helpers as ph

from .com import fs
from .com import db
from .com import sh


_UI_VIDEOSOUCE_JNODE_DIRPATH = '/etc/edge/setupstore/admin/actions/pipeline/%s/videosource'
_UI_EVENTSTORE_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/eventstore.json'
_UI_EVENTSTORE_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/eventstore-jnode.json'
_UI_PIPELINE_SET_FILEPATH = '/etc/edge/setupstore/admin/actions/pipeline.json'
_WS_RESTART_FILEPATH = '/var/spool/edge/ws0-ws0-restart'
_SS_RESTART_FILEPATH = '/var/spool/edge/ss0-ss0-restart'
PIPELINE_SET_FILEPATH = '/etc/edge/setupstore/admin/actions/pipeline.json'
PIPELINE_DEFSET_FILEPATH = '/opt/edge/share/setupstore/ui/pipeline-jnode.json'
EDGE_EVENTSTORE_DIRPATH = '/etc/edge/eventstore.65535'

_RUNTIME_CRON_FILEPATH = '/etc/cron.d/edge-base'
_DEFAULT_CRON_FILEPATH = '/opt/edge/share/cron.d/edge-base'


def restart_edge_module(edge_type, edge_id):
    """
    :type edge_type: str
    :type edge_id: str
    """
    eventlog.log_i('exec: {}.{}'.format(edge_type, edge_id))
    sh.run_silent("kill -9 `ps ax |"
                  "grep \"%s --id.%s\" |"
                  "grep -v grep |"
                  "awk '{print $1}'`" % (edge_type, edge_id))


def sighup_edge_module(edge_type, edge_id):
    """
    :type edge_type: str
    :type edge_id: str
    """
    eventlog.log_i('exec: {}.{}'.format(edge_type, edge_id))
    sh.run_silent("kill -1 `ps ax |"
                  "grep \"%s --id.%s\" |"
                  "grep -v grep |"
                  "awk '{print $1}'`" % (edge_type, edge_id))


def restart_webserver_lazy():
    eventlog.log_d('restarting webserver')
    # PCD restarts webserver when this file appears
    silent_mkdir_r(os.path.dirname(_WS_RESTART_FILEPATH))
    ws_restart_filepath = _WS_RESTART_FILEPATH

    with open(ws_restart_filepath, 'w'):
        pass


def get_edge_group(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return str(int(ui_pid) * 10)


def get_edge_id(edge_group, edge_identity):
    """
    :type edge_group: str
    :type edge_identity: EdgeIdentity
    :rtype: str
    """
    if edge_identity.type in [ph.EDGE_TYPE_WEBSERVER, ph.EDGE_TYPE_SETUPSTORE]:
        return '0'

    elif edge_identity.type in [ph.EDGE_TYPE_EVENTSTORE]:
        return '65535'

    return str(int(edge_group) + int(edge_identity.gid))


def restart_edge_modules(edge_type):
    """
    :type edge_type: str
    """
    eventlog.log_d('restarting all edge modules %s' % edge_type)
    subprocess.Popen(["pkill -9 %s" % edge_type], shell=True)


def restart_edge_server():
    # PCD restarts edgeserver when this file appears
    with open(_SS_RESTART_FILEPATH, 'w'):
        pass


def sighup_setupstore():
    eventlog.log_d('restarting setupstore')
    os.kill(int(subprocess.check_output(['pidof', 'setupstore'])), signal.SIGHUP)


def restart_eventstore():
    eventlog.log_d('restarting eventstore')
    subprocess.Popen(["pkill -9 eventstore"], shell=True)


def silent_mkdir_r(target_path):
    """
    :type target_path: str
    """
    if not target_path:
        return

    try:
        os.makedirs(target_path)

    except (OSError, TypeError) as x:
        eventlog.log_d("fail: make %s (%s)" % (target_path, x.message))


def silent_rm_r(target_path):
    """
    :type target_path: str
    """
    if not target_path:
        return

    try:
        # try remove file
        try:
            os.remove(target_path)
            return
        except EnvironmentError:
            pass

        # try remove nested-files in directory
        shutil.rmtree(target_path, ignore_errors=True)

    except EnvironmentError as x:
        eventlog.log_d("fail: remove %s (%s)" % (target_path, x.message))


def get_videsource_status(ui_pid, ui_id):
    """
    :type ui_pid: str
    :type ui_id: str
    :rtype: int | None
    """
    ui_jnode_path = os.path.join(_UI_VIDEOSOUCE_JNODE_DIRPATH % ui_pid, "0.json")

    ui_jnode = fs.load_json(ui_jnode_path)
    if ui_jnode is None:
        return None

    stats_dirpath = ui_jnode['ro']['stats-dirpath']
    stats_filepath = os.path.join(stats_dirpath, 'pipeline-%s/zabbix' % ui_pid, 'vs%s' % ui_id)

    stat_jnode = fs.load_json(stats_filepath)
    if stat_jnode is None:
        return None
    else:
        return stat_jnode['stream']['error']


def get_rn_image_wh_dh(ui_pid, ui_id):
    """
    :type ui_pid: str
    :type ui_id: str
    :rtype: (int, int, int) | None
    """
    stats_path = os.path.join('/var/spool/edge/pipeline-%s/zabbix/or%s' % (ui_pid, ui_id))
    stats_jnode = fs.load_json(stats_path)
    if not stats_jnode:
        return

    jnode = stats_jnode.get('stream')
    if not jnode:
        return

    image_wh = jnode.get('wh')
    image_dh = jnode.get('dh')
    if None not in [image_wh, image_dh]:
        return image_wh[0], image_wh[1], image_dh

    eventlog.log_i('fail: parse {}'.format(stats_path))


def get_videosource_image_wh(ui_pid, ui_id, ui_vs_jnode=None):
    """
    :type ui_pid: str
    :type ui_id: str
    :type ui_vs_jnode: dict | None
    :rtype: (int, int) | None
    """
    if ui_vs_jnode is None:
        ui_vs_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, ph.UI_TYPE_VIDEOSOURCE, ui_id)
        ui_vs_jnode = fs.load_json2(ui_vs_path)

    ro_jnode = ui_vs_jnode.get('ro', dict())
    stats_dir = ro_jnode.get('stats-dirpath')
    if not stats_dir:
        eventlog.log_i('fail: param #/ro/stats-dirpath has invalid value')
        return None

    stats_file = os.path.join(stats_dir, 'pipeline-%s/zabbix' % ui_pid, 'vs%s' % ui_id)
    stats_jnode = fs.load_json(stats_file)
    if stats_jnode is None:
        return None

    if stats_jnode['stream']['error']:
        eventlog.log_d('stat-file error: {}'.format(stats_jnode['stream']['error']))

    try:
        image_wh = stats_jnode['stream']['wh']
        return image_wh
    except KeyError:
        eventlog.log_e('fail: parse {}'.format(stats_file))


def get_pipeline_jnode_set():
    """
    :rtype: list[dict] | None
    """
    # create ui-jnode
    pipeline_jnode_set = get_pipeline_jnode_set2()
    if pipeline_jnode_set is not None:
        for pipeline in pipeline_jnode_set:
            pipeline.pop('type', None)

        pipeline_jnode_set.sort(key=lambda x: x['id'])
        return pipeline_jnode_set

    return None


def get_pipeline_jnode_set2():
    """
    :rtype: list[dict] | None
    """
    # create ui-jnode
    if not os.path.exists(PIPELINE_SET_FILEPATH):
        # noinspection PyBroadException
        try:
            fs.create_jnode_from_default(PIPELINE_DEFSET_FILEPATH, PIPELINE_SET_FILEPATH)

        except Exception:
            eventlog.log_x()
            return None

    with open(PIPELINE_SET_FILEPATH, 'r') as json_file:
        origins = ujson.load(json_file)

    return origins


def set_ui_pipeline_eventstore_jnode(jnode):
    """
    :rtype: dict | None
    """
    fs.write_json_file(jnode, _UI_EVENTSTORE_JNODE_FILEPATH)


def get_ui_pipeline_eventstore_jnode():
    """
    :rtype: dict | None
    """
    # create ui-jnode from default
    if not os.path.exists(_UI_EVENTSTORE_JNODE_FILEPATH):
        try:
            fs.create_jnode_from_default(_UI_EVENTSTORE_DEFJNODE_FILEPATH, _UI_EVENTSTORE_JNODE_FILEPATH)
        except RuntimeError as x:
            eventlog.log_d('fail: create jnode from def (%s)' % x)
            return None

    ui_jnode = fs.load_json(_UI_EVENTSTORE_JNODE_FILEPATH)
    return ui_jnode


def get_detector_limits(ui_pid, videosource_ui_id, vehicle=True, scene=True, lanes=False):
    """
    :type ui_pid: str
    :type videosource_ui_id: str
    :type vehicle: bool
    :type scene: bool
    :type lanes: bool
    :rtype: dict
    """
    ui_eventstore_jnode = get_ui_pipeline_eventstore_jnode()

    if ui_eventstore_jnode is None:
        ui_eventstore_jnode = list()

    wh = get_videosource_image_wh(ui_pid, videosource_ui_id)
    w, h = 640, 480
    if wh is not None:
        w, h = wh

    limits_jnode = {
        'images': {
        },
        'stores': ui_eventstore_jnode
    }
    if vehicle:
        limits_jnode['images']['vehicle'] = {'max-width': 640}

    if scene:
        limits_jnode['images']['scene'] = {'max-width': w}

    if lanes:
        ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(
            ui_pid, ph.UI_TYPE_LANE_ENFORCEMENT, videosource_ui_id)

        ui_jnode = fs.load_json2(ui_jnode_path)
        lanes_jnode = ui_jnode['rw']['args']['lanes']
        limits_jnode['lanes'] = map(lambda x: x['name'], lanes_jnode)

    return limits_jnode


def get_origin_uuid(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    pipeline_jnode_set = get_pipeline_jnode_set()
    if pipeline_jnode_set is None:
        eventlog.log_d('fail: get origin uuid from pipeline jnode set')
        return ''

    uuid = None
    for pipeline_jnode in pipeline_jnode_set:
        if ui_pid == pipeline_jnode['id']:
            uuid = pipeline_jnode['uuid']

    return uuid if uuid else ''


def get_origin_pid(search_uuid):
    """
    :type search_uuid: str
    :rtype: str | None
    """
    pipeline_jnode_set = get_pipeline_jnode_set()
    if pipeline_jnode_set is None:
        eventlog.log_d('fail: get origin pid from pipeline jnode set')
        return None

    pipeline_jnode_list = filter(lambda x: x['uuid'] == search_uuid, pipeline_jnode_set)

    pid = None
    if len(pipeline_jnode_list) == 1:
        pid = pipeline_jnode_list[0]['id']

    return pid


def get_origin_pids_set():
    """
    :rtype: set[str]
    """
    pipeline_jnode_set = get_pipeline_jnode_set()
    if pipeline_jnode_set is None:
        eventlog.log_d('fail: get origin pid from pipeline jnode set')
        return None

    pid_list = map(lambda x: x['id'], pipeline_jnode_set)
    return set(pid_list)


def get_origin_name(ui_pid):
    """
    :type ui_pid: str
    :rtype: str | None
    """
    ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, ph.UI_TYPE_EVENTSTORE, '0')
    ui_jnode = fs.load_json(ui_jnode_path)
    if ui_jnode:
        ui_jnode_rw = ui_jnode.get('rw') or dict()
        return ui_jnode_rw.get('origin-name', ui_pid)

    return ui_pid


def load_pscuctl_json():
    """
    :rtype: dict | None
    """
    psuctl_conf = fs.load_json(ph.PSUCTL_JSON_FILEPATH)
    return psuctl_conf


def get_psuctl_endpoint():
    """
    :rtype: str | None
    """
    psuctl_conf = load_pscuctl_json()
    if not psuctl_conf:
        eventlog.log_d('fail: psu-ctl.conf not found')
        return

    ro_jnode = psuctl_conf.get('ro')
    if not ro_jnode:
        eventlog.log_d('fail: param psu-ctl.conf#/ro has invalid value')
        return

    com_port = ro_jnode.get('com-port')
    if not com_port:
        eventlog.log_d('fail: param psu-ctl.conf#/ro/com-port has invalid value')
        return

    return com_port


def edit_db_origin(uuid, name='', media=True, manual=True):
    """
    :type uuid: str
    :type name: str
    :type media: bool
    :type manual: bool
    """
    dbapi_query_jnode = {
        'uuid': uuid,
        'name': name,
        'param': {
            'public': {
                'uuid': uuid,
                'name': name,
                'services': {
                    'event': ['feed', 'arch', 'stats']
                }
            },
            'manual': manual
        }
    }
    if media:
        dbapi_query_jnode['param']['public']['services']['media'] = ['live']

    db_args = db.DbArgs.deserialize2()

    def db_handler(db_link):
        """
        :rtype: dict
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "select errcode, errmsg, result from api_origin_set(%s)", [db.DbJsonArg(dbapi_query_jnode)])

            errcode, errmsg, api_reply_jnode = db_cursor.fetchone()
            if errcode != 0:
                eventlog.log_d('fail: api_origin_set (%s)' % errmsg)
                raise db.DbError

            return

    try:
        return db.with_db(db_args, db_handler)

    except db.DbError:
        eventlog.log_d('fail: db')
        raise db.DbError


def make_user_jnode(user_name, passwd, user_role):
    """
    :type user_name: str
    :type passwd: str
    :type user_role: str
    :raises NotImplemented:
    :rtype: dict
    """
    openssl = subprocess.Popen("openssl passwd -apr1 %s" % passwd, shell=True, stdout=subprocess.PIPE)

    new_password = openssl.communicate()[0]
    salt_method = new_password.split('$')[1]
    salt_value = new_password.split('$')[2]
    passwd_hash = new_password.split('$')[3].split('\n')[0]

    if user_role == 'user':
        jnode = {
            "section": "users",
            "key": user_name,
            "value": {
                "system": {
                    "salt": [salt_method, salt_value],
                    "hash": passwd_hash,
                    "role": ["user"]
                },
                "public": {
                    "name": user_name
                }
            }
        }
    elif user_role == 'admin':
        jnode = {
            "section": "users",
            "key": user_name,
            "value": {
                "system": {
                    "salt": [salt_method, salt_value],
                    "hash": passwd_hash,
                    "role": ["user", "admin"]
                },
                "public": {
                    "name": user_name,
                    "rights": {
                        "archives": True,
                        "modules": {
                            "system": [
                                "update",
                                "zabbix"
                            ],
                            "origins": True
                        },
                        "detectors": True
                    }
                }
            }
        }

    else:
        raise NotImplemented

    return jnode


def get_detector_types(ptype):
    """
    :type ptype: str
    :rtype: list[str]
    """
    detectors_types = ['parking', 'wrong-way', 'red-light', 'wanted', 'section-speed']
    if ptype == ph.PTYPE_SPEED:
        detectors_types.append('spot-speed')

    return detectors_types


def change_eventstore_systest(systest):
    """
    :param systest: bool
    """
    edge_jnode_dir = EDGE_EVENTSTORE_DIRPATH
    main_conf_abspath = os.path.join(edge_jnode_dir, 'main.conf')
    main_conf_jnode = fs.load_json(main_conf_abspath)
    main_conf_jnode['systest'] = systest
    fs.write_json_file(main_conf_jnode, main_conf_abspath)

    origins_jnode = get_pipeline_jnode_set2()

    for origin_jnode in origins_jnode:
        ui_pid = origin_jnode['id']
        edge_group = get_edge_group(ui_pid)
        edge_identity = ph.EdgeIdentity(ph.EDGE_TYPE_SPEEDCONTROL, edge_group)
        edge_id = get_edge_id(edge_group, edge_identity)
        jnode_dirpath = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_SPEEDCONTROL, edge_id)
        jnode_filepath = fs.join_path(jnode_dirpath, 'transit.conf')
        jnode = fs.load_json(jnode_filepath)
        jnode['systest'] = systest
        fs.write_json_file(jnode, jnode_filepath, True)


def get_es_uds_endpoint():
    """
    :rtype: str
    """
    return '/var/spool/edge/system/es-es-uds'


def get_ra_origin_pid():
    """
    :rtype: str | None
    """
    origins_jnode = get_pipeline_jnode_set2()
    ui_pid = None
    for origin_jnode in origins_jnode:
        if origin_jnode['type'] == ph.PTYPE_SPEED:
            if ui_pid is not None:
                eventlog.log_i('fail: multiple radar origins')
                return None
            ui_pid = origin_jnode['id']

    return ui_pid


def get_ra_type(ui_pid):
    """
    :rtype: str | None
    """
    ui_jnode_path = ph.get_ui_pipeline_jnode_filepath(ui_pid, ph.UI_TYPE_LOCATION, '0')
    ui_jnode = fs.load_json(ui_jnode_path)
    if ui_jnode:
        ui_jnode_ro = ui_jnode.get('ro') or dict()
        return ui_jnode_ro.get('type')


def mk_origin_about(ui_pid, uuid, time=None, place=None):
    """
    :type ui_pid: str
    :type uuid: str
    :type time: str | None
    :type place: str | None
    :return:
    """
    conf_path = ph.get_origin_about_filepath(ui_pid)
    jnode = fs.load_json(conf_path)

    if jnode is None:
        jnode = {
            "origin": {
                "uuid": uuid
            }
        }
    else:
        jnode['origin']['uuid'] = uuid

    if time:
        jnode['origin']['time'] = time

    if place:
        jnode['location'] = {
            'place': place
        }

    fs.write_json_file(jnode, conf_path)


def disable_edge_cron():
    try:
        os.remove(_RUNTIME_CRON_FILEPATH)
    except OSError as x:
        if x.errno == errno.ENOENT:
            eventlog.log_d('no cron-file found')
        else:
            raise


def enable_edge_cron():
    shutil.copy(_DEFAULT_CRON_FILEPATH, _RUNTIME_CRON_FILEPATH)
